"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * This module handles the concerns of publishing. It handles interaction with
 * git tagging, pushing to the git origin, the package registry, etc.
 */
const fs = __importStar(require("fs"));
const Path = __importStar(require("path"));
const proc = __importStar(require("./proc"));
const promise_1 = __importDefault(require("simple-git/promise"));
const utils_1 = require("./utils");
const defaultOpts = {
    vPrefix: false,
};
/**
 * Run the publishing process.
 *
 * 1. Change package.json version field to be new version.
 * 2. npm publish --tag next.
 * 3. discard package.json change.
 * 4. git tag {newVer}.
 * 5. git tag next.
 * 6. git push --tags.
 *
 */
function publish(release, givenOpts) {
    var _a;
    return __awaiter(this, void 0, void 0, function* () {
        const opts = Object.assign(Object.assign({}, defaultOpts), givenOpts);
        // Update package.json
        const packageJsonPath = Path.join(process.cwd(), 'package.json');
        const packageJsonString = yield fs.readFileSync(packageJsonPath, {
            encoding: 'utf8',
        });
        const packageJson = JSON.parse(packageJsonString);
        if (!packageJson) {
            throw new Error(`Looked for but could not find a package.json file at ${packageJsonPath}. This file is required for publishing to the npm registry.`);
        }
        if (typeof packageJson !== 'object') {
            throw new Error(`Found a package.json file at ${packageJsonPath} but it appears to be malformed. It did not parse into an object but rather: ${packageJson}`);
        }
        const updatedPackageJson = Object.assign(Object.assign({}, packageJson), { version: release.version });
        yield fs.writeFileSync(packageJsonPath, JSON.stringify(updatedPackageJson));
        console.log('updated package.json in prep for publishing');
        // publish to the npm registry
        // If we are using a script runner then publish with that same tool. Otherwise
        // default to using npm. The reason we need to do this is that problems occur
        // when mixing tools. For example `yarn run ...` will lead to a spawn of `npm
        // publish` failing due to an authentication error.
        const scriptRunner = (_a = utils_1.detectScriptRunner(), (_a !== null && _a !== void 0 ? _a : 'npm'));
        const runPublishString = scriptRunner === 'npm'
            ? `npm publish --tag ${release.distTag}`
            : scriptRunner === 'yarn'
                ? `${scriptRunner} publish --tag ${release.distTag} --no-git-tag-version --new-version ${release.version}`
                : utils_1.casesHandled(scriptRunner);
        yield proc.run(runPublishString, { require: true });
        console.log('published package to the npm registry');
        if (release.additiomalDistTags) {
            for (const distTag of release.additiomalDistTags) {
                const runString = scriptRunner === 'npm'
                    ? `npm dist-tags add ${packageJson.name}@${release.version} ${distTag}`
                    : scriptRunner === 'yarn'
                        ? `yarn tag add ${packageJson.name}@${release.version} ${distTag}`
                        : utils_1.casesHandled(scriptRunner);
                yield proc.run(runString, { require: true });
                console.log(`updated dist-tag "${distTag}" to point at this version`);
            }
        }
        const git = promise_1.default();
        // TODO no invariant in system that checks that package.json was not modified
        // before beginning the publishing process. In other words we may be losing
        // user work here. This check should be in strict mode.
        yield git.checkout('package.json');
        console.log('reverted package.json changes now that publishing is done');
        // tag the commit
        const versionTag = opts.vPrefix ? 'v' + release.version : release.version;
        yield git.addAnnotatedTag(versionTag, versionTag);
        console.log(`tagged this commit with ${versionTag}`);
        yield git.pushTags();
        console.log(`pushed tag to remote`);
    });
}
exports.publish = publish;
